/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// refresh.h -- public interface to refresh functions

#define	MAXCLIPPLANES	11

#define	TOP_RANGE		16			// soldier uniform colors
#define	BOTTOM_RANGE	96

//=============================================================================

typedef struct efrag_s
{
	struct mleaf_s		*leaf;
	struct efrag_s		*leafnext;
	struct entity_s		*entity;
	struct efrag_s		*entnext;
} efrag_t;


#define LERP_MOVESTEP	(1 << 0) //this is a MOVETYPE_STEP entity, enable movement lerp
#define LERP_RESETANIM	(1 << 1) //disable anim lerping until next anim frame
#define LERP_RESETANIM2	(1 << 2) //set this and previous flag to disable anim lerping for two anim frames
#define LERP_RESETMOVE	(1 << 3) //disable movement lerping until next origin/angles change
#define LERP_FINISH		(1 << 4) //use lerpfinish time from server update instead of assuming interval of 0.1

typedef struct entity_s
{
	glmatrix				localmatrix;

	qboolean				bbculled;

	// textures for MDLs
	int						gl_texturenum;
	int						fb_texturenum;

	qboolean				forcelink;		// model changed

	int						update_type;

	entity_state_t			baseline;		// to fill in defaults in updates

	double					msgtime;		// time of last update
	vec3_t					msg_origins[2];	// last two updates (0 is newest)	
	vec3_t					origin;
	vec3_t					msg_angles[2];	// last two updates (0 is newest)
	vec3_t					angles;	
	struct model_s			*model;			// NULL = no model
	struct efrag_s			*efrag;			// linked list of efrags
	int						frame;
	float					syncbase;		// for client-side animations
	byte					*colormap;
	int						effects;		// light, particals, etc
	int						skinnum;		// for Alias models
	int						visframe;		// last frame this entity was
											//  found in an active leaf

	// rockettrails
	vec3_t					oldtrailorigin;
	double					nexttrailtime;

	int						lerpflags;		// lerping

	float					lerpstart;		// animation lerping
	float					lerptime;		// animation lerping
	short					previouspose;	// animation lerping
	short					currentpose;	// animation lerping
	float					lerpblend;		// blending factor between poses

	float					movelerpstart;	// transform lerping
	vec3_t					previousorigin;	// transform lerping
	vec3_t					currentorigin;	// transform lerping
	vec3_t					lerpedorigin;
	vec3_t					previousangles;	// transform lerping
	vec3_t					currentangles;	// transform lerping
	vec3_t					lerpedangles;
} entity_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct refdef_s
{
	// padded for shaders
	float		vieworigin[4];
	float		viewangles[4];
	float		fov_x, fov_y;
} refdef_t;


// refresh
extern	int		reinit_surfcache;


extern	refdef_t	r_refdef;
extern	float	vup[4];
extern	float	vpn[4];
extern	float	vright[4];
extern	float	r_origin[4];

extern	struct texture_s	*r_notexture_mip;


void R_Init (void);
void R_InitTextures (void);
void R_InitEfrags (void);
void R_RenderView (void);		// must set r_refdef first
void R_InitSky (byte *src);	// called at level load

void R_AddEfrags (entity_t *ent);

void R_AddVisedict (entity_t *e);

void R_NewMap (void);
void V_NewMap (void);

void R_ParseParticleEffect (void);
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count);
void R_RocketTrail (vec3_t start, vec3_t end, int type);

#ifdef QUAKE2
void R_DarkFieldParticles (entity_t *ent);
#endif
void R_EntityParticles (entity_t *ent);
void R_BlobExplosion (vec3_t org);
void R_ParticleExplosion (vec3_t org);
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength);
void R_LavaSplash (vec3_t org);
void R_TeleportSplash (vec3_t org);


// surface cache related
extern	int		reinit_surfcache;	// if 1, surface cache is currently empty and
extern qboolean	r_cache_thrash;	// set if thrashing the surface cache

int	D_SurfaceCacheForRes (int width, int height);
void D_DeleteSurfaceCache (void);
void D_InitCaches (void *buffer, int size);


